@@ -1,6 +1,6 @@
 import numpy as np
 from itertools import tee
-from tensorflow.keras.models import model_from_yaml
+from tensorflow.keras.models import model_from_json
 from tensorflow.keras.optimizers import get as get_optimizer
 from tensorflow.python.keras.utils.generic_utils import slice_arrays
 
@@ -26,8 +26,9 @@ class SparkWorker(object):
     def train(self, data_iterator):
         """Train a keras model on a worker
         """
+        history = None
         optimizer = get_optimizer(self.master_optimizer)
-        self.model = model_from_yaml(self.yaml, self.custom_objects)
+        self.model = model_from_json(self.yaml, self.custom_objects)
         self.model.compile(optimizer=optimizer,
                            loss=self.master_loss, metrics=self.master_metrics)
         self.model.set_weights(self.parameters.value)
@@ -36,17 +37,16 @@ class SparkWorker(object):
         x_train = np.asarray([x for x, y in feature_iterator])
         y_train = np.asarray([y for x, y in label_iterator])
 
-        self.model.compile(optimizer=get_optimizer(self.master_optimizer),
-                           loss=self.master_loss,
-                           metrics=self.master_metrics)
-
         weights_before_training = self.model.get_weights()
         if x_train.shape[0] > self.train_config.get('batch_size'):
-            self.model.fit(x_train, y_train, **self.train_config)
+            history = self.model.fit(x_train, y_train, **self.train_config)
         weights_after_training = self.model.get_weights()
         deltas = subtract_params(
             weights_before_training, weights_after_training)
-        yield deltas
+        if history:
+            yield [deltas, history.history]
+        else:
+            yield [deltas, None]
 
 
 class AsynchronousSparkWorker(object):
@@ -84,7 +84,7 @@ class AsynchronousSparkWorker(object):
         if x_train.size == 0:
             return
 
-        self.model = model_from_yaml(self.yaml, self.custom_objects)
+        self.model = model_from_json(self.yaml, self.custom_objects)
         self.model.compile(optimizer=get_optimizer(self.master_optimizer),
                            loss=self.master_loss, metrics=self.master_metrics)
         self.model.set_weights(self.parameters.value)
